import numpy as np

class Model():
    
    def __init__(self, layers):
        self.layers = layers

    def __call__(self, x):
        return self.forward(x)

    def forward(self, x):
        for layer in self.layers:
            y = layer(x)
            x = y
        return y
    
    def backward(self, dL):
        for layer in reversed(self.layers):
            dL = layer.backward(dL)
        return dL
    
    def optimize(self):
        for layer in self.layers:
            layer.optimize()

    def set_optimizer(self, optimizer):
        for layer in self.layers:
            layer.set_optimizer(optimizer)

    def set_loss(self, loss_fn):
        self.loss_fn = loss_fn

    def setup(self, loss_fn, optimizer):
        self.set_loss(loss_fn)
        self.set_optimizer(optimizer)

    def train(self, x_train, y_train, x_test=None, y_test=None, split=0.2, epochs=100, batch_size=10, lr=5e-3, verbose=1):
        steps = x_train.shape[0]//batch_size
        loss,accuracy = [],[]
        for epoch in range(epochs):
            epoch_loss = 0
            for i in range(0, x_train.shape[0], batch_size):
                x, y = x_train[i:i+batch_size], y_train[i:i+batch_size]
                y_pred = self.forward(x)

                epoch_loss += self.loss_fn(y, y_pred)
                
                dL_dy = self.loss_fn.gradient()

                self.backward(dL_dy)

                self.optimize()
            epoch_loss_test = epoch_loss/steps
            loss.append(epoch_loss_test)
            if x_test is not None:
                accuracy.append(self.evaluate_accuracy(x_test, y_test))
            
            if verbose == 1:
                print("Epoch {} training loss = {}".format(epoch, epoch_loss_test) )
        self.loss = loss
        self.accuracy = accuracy
        return loss,accuracy

    def predict(self, x):
        estimate = self.forward(x)
        return estimate

    def evaluate_accuracy(self, x_val, y_val):
        num_corrects = 0
        for i in range(x_val.shape[0]):
            if self.predict(x_val) == y_val:
                num_corrects += 1
        return num_corrects/x_val.shape[0]


class Layer:
    '''
    Abstract class with Layer definitions
    '''

    def __init__(self):
        pass

    def forward(self, x):
        pass

    def __call__(self, x):
        return self.forward(x)

    def backward(self, dL_dy):
        pass

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def optimize(self):
        pass


import numpy as np
from ..layers import Layer

class Sigmoid(Layer):
    '''
    Learn more about Sigmoid Function here: https://en.wikipedia.org/wiki/Sigmoid_function
    '''

    def forward(self, x):
        self.last_x = x
        self.result = 1.0 / (1.0 + np.exp(-x))
        return self.result

    def __call__(self, x):
        return self.forward(x)

    def backward(self, dL_dy):
        self.dy_dx = self.result * (1 - self.result)
        return self.dy_dx * dL_dy

class sigmoid(Layer):
    def __init__(self):
        pass


import numpy as np
from ..layers import Layer

class ReLU(Layer):
    '''
    '''

    def forward(self, x):
        self.last_x = x
        self.result = np.maximum(0, x)
        return self.result

    def __call__(self, x):
        return self.forward(x)

    def backward(self, dL_dy):
        self.dy_dx = self.last_x
        self.dy_dx[self.dy_dx <= 0 ] = 0
        self.dy_dx[self.dy_dx > 0] = 1
        return self.dy_dx * dL_dy

class relu(ReLU):
    def __init__(self):
        pass


import numpy as np


class MeanAbsoluteError():

    def __call__(self, y, y_pred):
        self.last_y_pred = y_pred
        self.last_y = y

        assert y_pred.shape == y.shape
        sums = np.sum(np.abs( y_pred - y ), axis=0)
        n = y.shape[0]
        
        self.last_loss = sums/n        
        return self.last_loss

    def gradient(self):
        self.dL_dy = self.last_y_pred > self.last_y
        self.dL_dy = self.dL_dy.astype(int)
        self.dL_dy[ self.dL_dy == 0 ] = -1 
        return self.dL_dy

class MAE(MeanAbsoluteError):

    def __init__(self):
        pass


from numpy_nn.losses import MAE
import numpy as np

def test_mae_basic1():
    y_pred = np.array([[1,2],[3,4]])
    y = np.array([[2,3],[4,5]])
    mae = MAE()
    true_list = mae(y,y_pred)
    assert list(true_list) == [True, True]


import numpy as np
from numpy_nn import Neuron
from numpy_nn.losses import MAE

def test_neuron_forward_operation():
    x = np.array([[1,1,1,1]])

    # create single neuron
    neuron = Neuron(4, 2)

    result = neuron.forward(x)
    print(result)
    assert result.shape == (1, 2)


import numpy as np


class MeanSquaredError():

    def __call__(self, y, y_pred):
        self.last_y_pred = y_pred
        self.last_y = y

        assert y_pred.shape == y.shape
        self.last_loss = np.sum(np.square(y-y_pred), axis=0)/y_pred.shape[0]
        return self.last_loss

    def gradient(self):
        self.dL_dy = -2*(self.last_y - self.last_y_pred)/self.last_y.shape[0]
        return self.dL_dy

class MSE(MeanSquaredError):
    def __init__(self):
        pass


"""Top-level package for numpy_nn."""

__author__ = """Rahul D Shetty"""
__email__ = '35rahuldshetty@gmail.com'
__version__ = '0.1.0'

# Layers
from .layers import Neuron

# Losses
from .losses import MeanAbsoluteError



'''
This module defines a layer of fully connected Neurons.
'''
import numpy as np
from .Layer import Layer

class Neuron(Layer):
    '''
    A Basic layer of neurons which fully connects to the next layer.
    '''
    def __init__(self, num_inputs, num_outputs):
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs

        self.weights = np.random.standard_normal((num_inputs, num_outputs))
        self.biases = np.random.standard_normal(num_outputs)

    def forward(self, x):
        self.last_input = x
        self.result = np.dot(x, self.weights) + self.biases
        return self.result

    def __call__(self, x):
        return self.forward(x)

    def backward(self, dL_dy):
        dy_db = np.ones(dL_dy.shape[0]).astype('float32')
        dy_dw = self.last_input.T
        dy_dx = self.weights.T

        self.dL_db = np.dot(dy_db, dL_dy)
        self.dL_dw = np.dot(dy_dw, dL_dy)

        self.dL_dx = np.dot(dL_dy, dy_dx)

        return self.dL_dx

    def optimize(self):
        self.weights = self.optimizer(self.weights, self.dL_dw)
        self.biases = self.optimizer(self.biases, self.dL_db)



import numpy as np
from ..layers import Layer

class TanH(Layer):
    '''
    '''
    def forward(self, x):
        self.last_x = x
        self.result = (np.exp(x) - np.exp(-x))/(np.exp(x) + np.exp(-x))
        return self.result

    def __call__(self, x):
        return self.forward(x)

    def backward(self, dL_dy):
        self.dy_dx = 1 - self.result**2
        return self.dy_dx * dL_dy

class tanh(TanH):
    def __init__(self):
        pass
